/*
  GPL Driver for the Codec Interface on MGA G1/2/400 cards
  (requires mgacore.c)

 * (c) 1999-2000 David Barth <dbarth@besancon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Publi License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *****************************************************************************
 *
 * Modified and extended by
 *      Mike Bernson <mike@mlb.org>
 *	Merrick Johnson <inkling@nop.org>
 *	Leon van Stuivenberg <leonvs@iae.nl>
 */

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include "mgacore.h"
#include "mgacodec.h"

#define dprintk if (mgacore_debug) printk
#define dprintk2 if (mgacore_debug>1) printk


/* TODO :
   - iovec inputs ?
*/

/****************************************************************************
 * MATROX CODEC INTERFACE                                                   *
 ****************************************************************************/

/****************************************************************************
 * PRIVATE SECTION : implementations details                                *
 *                                                                          *
 * (see the end of the file for public functions implementation)            *
 *                                                                          *
 ****************************************************************************/

/*
 * Private structure definition. Only the codec module has to know the
 * details. Modules using the codec, should access it through the
 * public API.
 */
struct mga_codec_s
{
	/* address of the codec interface inside the framebuffer */
        u32 addr;

	/* codec interface mode (VMI-A/B or I33) */
	int mode;

	/* value to xor with the register addresses */
	int xor_regs;

	/* register (in the external chip) to send bitstream to */
	int dstreg;

	/* recovery width, see G200 spec, p 4-89 for timing information */
	int rwidth;

	/* shadow registers for CODECCTL:
	 * these are Write-Only registers on the G200, so this copy lets
	 * us read it to see what the latest settings are
	 * NOTE: CODECCTL+2, bits 16-23 of CODECCTL are unused.
	 * WARNING: CODECCTL register should ONLY be programmed byte at a time
	 * -inkling
	 */
	u8 misc_shadow;		/* CODECCTL+3, miscctl byte */
	u8 rwcfa_shadow;	/* CODECCTL+1, recovery width, codecfifoaddr */
	u8 ctl_shadow;		/* CODECCTL+0, general codec control bits */
	spinlock_t lock;	/* lock for xx_shadow */

	/* buffers in host memory holding a copy of the command and data
	   buffer of the codec interface.  The command buffer is copied
	   at execution time to the framebuffer. Once the commands have
	   been executed the databuffer is filled with the data retrieved
	   from the codec and waiting in the framebuffer.
	*/
	u32 cmdbuffer[128]; /* organized as groups of 8 bytes (map for 512) */
	int cmdindex;
	int last_cmdindex;
	u16 databuffer[256]; /* organized as groups of 8 bytes */
	int dataindex;
	int readindex;
	int last_readindex;
	
	spinlock_t cmd_lock;

	/* see mgacodec_send_bitstream */
	u32 streamindex;

	/*
	 * transfer indicates a codec transfer has started and also which 
	 * direction it is going. defined values are (numeric):
	 *  CODEC_NO_TRANSFER	(0)	no transfer has started yet
	 *  CODEC_SEND_TRANSFER	(1)	data FROM G200 to the codec
	 *  CODEC_RECV_TRANSFER	(2)	data TO G200 from the codec
	 *
	 * stalled indicates a buffer level IRQ has stalled the transfer
	 * 
	 */
	int transfer;
	int stalled;
	spinlock_t state_lock; /* for transfer and stalled */

	wait_queue_head_t outq;
	wait_queue_head_t inq;
	wait_queue_head_t cmdq;
	wait_queue_head_t vsyncq;

	int cmdq_hascmpl;
	int vsyncq_hasirq;

	void	(*callback)(struct mga_dev_s *mga, int state);

	struct mga_dev_s *mga;
	
	int send_busy;	
};


#define CODEC_NO_TRANSFER	0
#define CODEC_SEND_TRANSFER	1
#define CODEC_RECV_TRANSFER	2

static void mgacodec_buflevel_irq(struct mga_dev_s *mga, struct pt_regs *regs);
static void mgacodec_cmdcmpl(struct mga_dev_s *mga, struct pt_regs *regs);
static void mgacodec_vsync_irq(struct mga_dev_s *mga, struct pt_regs *regs);

/* CODEC registers */
#define CODECCTL        0x3e40
#define CODECMISCCTL	0x3e43
#define CODECADDR       0x3e44
#define CODECHOSTPTR    0x3e48
#define CODECHARDPTR    0x3e4c
#define CODECLCODE      0x3e50


#define CODEC_CMDAREA_SIZE	0x0200
#define CODEC_ARGAREA_SIZE	0x0200


/*
 * VSTATUS bits
	not all of these are used by the 36710 (mgadvd)
	i've indicated the ones that are with ++ and the ones that should
	be with +-. other modules, i.e. RRG, will use a different subset
	see G200 spec, p 3-198 
	-inkling
 */
#define vinvsyncpen	0x0001	/*    video in vertical sync pending */
#define cmdcmplpen	0x0002	/* +- command complete interrupt pending */
#define blvlpen		0x0004	/* ++ buffer level interrupt pending */
#define dcmpeoipen	0x0008	/*    decomp end of image interrupt pending */
#define vinfielddetd	0x0100	/*    video input field detected */
#define vincapd		0x0200	/*    video in captured */
#define rawvbicapd	0x0400	/*    raw vbi captured */
#define slcvbicapd	0x0800	/*    slice vbi captured */
#define codec_stalled 	0x1000	/* ++ special use conditions */

/*
  codec_stalled is ONLY valid under certain conditions. see G200 spec p 3-199
  -inkling
*/

#define CODEC_WRITE(addr, byte) ( ((addr) << 8) | (byte) )


/*
 * wait some microseconds
 *
 */
static inline void mgacodec_wait(int delay)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(delay * HZ/1000000);
}

/* this is called by buffer level IRQ to stall the transfer when the hostptr
   limit IRQ has been activated because hardptr has reached hostptr.
   it will try to stall the transfer and set the stalled flag if it did.
   -inkling */
int
mgacodec_stall_transfer(struct mga_codec_s *codec)
{
	u8 ctl_mask = ~(codectransen);
	u8 ctl = 0;
	int i;
	unsigned long flags;

	spin_lock_irqsave( &codec->state_lock, flags );

	/* if transfer in progress and transfer not already stalled */
	if (codec->transfer) {
		if (!codec->stalled) {
			dprintk2("mgacodec: stalling %s transfer\n",
				(codec->transfer==CODEC_SEND_TRANSFER) ? "SEND":"RECEIVE");

			/* this will clear codectransen and cmdexectrig while
			   keeping all of the other ctl bits the same.
			   i.e. stop the transfer and dont trigger cmd exec.

			   FIXME: if we want to send commands to the zoran during
			   playback, we HAVE to use cmdexectrig during stall period.
			   to do this right will take some IRQ support for cmdcmplpen
			   to tell you when the command is complete and it's ok
			   to enable transfers again.
			   
			   this logic will have to be added because cmdexectrig by
			   itself would stop the transfer and trash current cmd
			   (and data?) fifo.  see G200 spec p 3-69.

			   setting cmdexectrig would clear codectransen and
			   that would NOT reflect in ctl_shadow without some logic.
			   BEST way might be to add a clear for codectransen if the
			   cmdexectrig is set in mgacodec_and_or_ctl to update shadow
			   at least from my interpretation of the G200 spec
			   -inkling */
			
			mgacodec_and_or_codecctl(codec, ctl_mask, ctl);

			for (i=160; i; i--) {
				udelay(1);
				if ((mga_read(codec->mga, VSTATUS) & codec_stalled)) {
					dprintk2( "mgacodec: stalled, left=%i\n", i );
					break;
				}
			}

			/* this is the only time codec_stalled status is valid */
			if (i==0) {
				printk("mgacodec: unable to stall %s transfer\n",
				       (codec->transfer==CODEC_SEND_TRANSFER)
				       ? "SEND":"RECEIVE");
				spin_unlock_irqrestore( &codec->state_lock, flags );
				return 1;
			}

			codec->stalled = 1;
			dprintk2("mgacodec: %s transfer stalled\n",
			  (codec->transfer==CODEC_SEND_TRANSFER)
			  ? "SEND":"RECEIVE");
		}
	}
	
	spin_unlock_irqrestore( &codec->state_lock, flags );
	return 0;
}

/* copy a block of data to the host framebuffer for transfer to the
   decoder chip.
*/
static inline void mgacodec_copy_bitstream(struct mga_codec_s *codec, const u8 *bitstream, int size)
{
	int chunk;
	u32 index;

	/*
	 * if size if greater than the room left at the end of the buffer,
	 * then we need to split the write operation in two. 1) fill the end of
	 * the buffer, 2) put the remaining bytes at beginning of the ring
	 * (thanks mike@mlb.org for finding it)
	 */

	index = codec->streamindex;

	chunk = CODEC_BUFFER_SIZE - index;
	if (size > chunk) {
		memcpy_toio((u8*)codec->mga->fb + codec->addr + index,
			bitstream, chunk);
		size -= chunk;
		bitstream += chunk;
		index = 0;
	}

	memcpy_toio((u8*)codec->mga->fb + codec->addr + index, bitstream, size);
	index += size;
	index %= CODEC_BUFFER_SIZE;
	
	mga_flush( codec->mga );
	
	codec->streamindex = index; /* automic/consistent update of streamindex */
}


/****************************************************************************
 * PUBLIC INTERFACE                                                         *
 *                                                                          *
 ****************************************************************************/


EXPORT_SYMBOL(mgacodec_attach);
EXPORT_SYMBOL(mgacodec_deattach);
EXPORT_SYMBOL(mgacodec_init);
EXPORT_SYMBOL(mgacodec_callback);
EXPORT_SYMBOL(mgacodec_reset);
EXPORT_SYMBOL(mgacodec_bitstream);
EXPORT_SYMBOL(mgacodec_and_or_codecctl);
EXPORT_SYMBOL(mgacodec_and_or_miscctl);
EXPORT_SYMBOL(mgacodec_set_misc2);
EXPORT_SYMBOL(mgacodec_set_ddc2);
EXPORT_SYMBOL(mgacodec_push_readb);
EXPORT_SYMBOL(mgacodec_push_readw);
EXPORT_SYMBOL(mgacodec_push_writeb);
EXPORT_SYMBOL(mgacodec_push_writew);
EXPORT_SYMBOL(mgacodec_push_pause);
EXPORT_SYMBOL(mgacodec_get_byte);
EXPORT_SYMBOL(mgacodec_get_word);
EXPORT_SYMBOL(mgacodec_exec_commands);
EXPORT_SYMBOL(mgacodec_get_transfer);
EXPORT_SYMBOL(mgacodec_stop_transfer);
EXPORT_SYMBOL(mgacodec_start_transfer);
EXPORT_SYMBOL(mgacodec_send_bitstream);
EXPORT_SYMBOL(mgacodec_send_bitstream_irq);
EXPORT_SYMBOL(mgacodec_recv_bitstream);
EXPORT_SYMBOL(mgacodec_buflevel_irq);
EXPORT_SYMBOL(mgacodec_buffer_free);
EXPORT_SYMBOL(mgacodec_set_transfer_limit);

EXPORT_SYMBOL(mgacodec_hardptr);
EXPORT_SYMBOL(mgacodec_lcode);

EXPORT_SYMBOL(mgacodec_begin_commands);
EXPORT_SYMBOL(mgacodec_end_commands);
EXPORT_SYMBOL(mgacodec_get_miscctl);
EXPORT_SYMBOL(mgacodec_stall_transfer);
EXPORT_SYMBOL(mgacodec_poll);

EXPORT_SYMBOL(mgacodec_is_stalled);

EXPORT_SYMBOL(mgacodec_vsync_sleep);

/**
 *	mga_codec_init - 	initialize the codec interface
 *	@mga:	mga instance
 *	@addr:	address of a work buffer
 *	@mode:	transfer mode (VMI or I33)
 *	@reg:	destination register for bitstream transfer
 *
 *	This function initializes and configures the Codec Interface
 *	for a particular target chipset. The Interface is set to
 *	IDLE state.
 *
 *	It is the responsability of the programmer to allocate a
 *	work buffer in video memory.
 *
 *	Transfer modes:
 *	%MGACODEC_MODE_I33
 *
 *	%MGACODEC_MODE_VMI_A
 *
 *	%MGACODEC_MODE_VMI_B
 **/
void mgacodec_attach(struct mga_dev_s *mga )
{
	struct mga_codec_s *codec;

	codec = kmalloc(sizeof(*codec), GFP_KERNEL);
	if (codec == NULL)
		return;

	memset(codec, 0, sizeof(*codec));

	/* p. 3-68 of the MGA HandBook : the Codec Addr must be aligned
	   on a 1Kb boundary
	*/
	codec->addr = mga_alloc_buffer(mga,
		CODEC_BUFFER_SIZE + CODEC_CMDAREA_SIZE + CODEC_ARGAREA_SIZE,
		1024 );

	codec->ctl_shadow = 0;
	codec->misc_shadow = 0;
	codec->cmdindex = 0;
	codec->last_cmdindex = 0;
	codec->dataindex = 0;
	codec->streamindex = 0;
	codec->transfer = 0;
	codec->readindex = 0;
	codec->last_readindex = 0;

	/* codec recovery width, play with this number to see what you get.
	   i did not notice any difference no matter what setting i used.
	   it may be more useful for the other encoder/decoders, and thus
	   should be part of mgacodec_init.
	   - inkling */
	codec->rwidth = 0;

	codec->mga = mga;
	//codec->mode =
	//codec->dstreg =
	//codec->xor_regs =
	//codec->rwcfa_shadow = ?
	codec->stalled = 0;
	codec->callback = NULL;
	
	spin_lock_init(&codec->cmd_lock);
	spin_lock_init(&codec->lock);
	spin_lock_init(&codec->state_lock);

	init_waitqueue_head(&codec->outq);
	init_waitqueue_head(&codec->inq);
	init_waitqueue_head(&codec->cmdq);
	init_waitqueue_head(&codec->vsyncq);

	mga->codec = codec;

	/* does this go here or at bottom of i34_attach ? -ink */
	/* doing it here makes it so you dont have to export vsync_irq */
	mga_enable_irq( mga, MGA_IRQ_VINVSYNC, (mga_interrupt_handler_t)mgacodec_vsync_irq);

	mga_enable_irq( mga, MGA_IRQ_CMDCMPL, (mga_interrupt_handler_t)mgacodec_cmdcmpl);
	mga_enable_irq( mga, MGA_IRQ_BUFLVL, (mga_interrupt_handler_t)mgacodec_buflevel_irq);
	printk("mgacodec: *mga @ %p\n",mga);
}


void
mgacodec_deattach(struct mga_codec_s *codec)
{
	if ( codec == NULL )
		return;
	mga_disable_irq( codec->mga, MGA_IRQ_BUFLVL, (mga_interrupt_handler_t)mgacodec_buflevel_irq);
	mga_disable_irq( codec->mga, MGA_IRQ_CMDCMPL, (mga_interrupt_handler_t)mgacodec_cmdcmpl);

	/* vsync sleeper off */
	mga_disable_irq( codec->mga, MGA_IRQ_VINVSYNC, (mga_interrupt_handler_t)mgacodec_vsync_irq);

	mgax_writeb(codec->mga, CODECCTL, 0);

	mga_free_buffer(codec->mga,
		CODEC_BUFFER_SIZE + CODEC_CMDAREA_SIZE + CODEC_ARGAREA_SIZE );

	codec->mga->codec = NULL;
	
	kfree(codec);
}


/**
 *	mgacodec_init - 	initialize the codec interface
 *	@codec:	codec instance
 *	@mode:	transfer mode (VMI or I33)
 *	@reg:	destination register for bitstream transfer
 *
 *	This function initializes and configures the Codec Interface
 *	for a particular target chipset. The Interface is set to
 *	IDLE state.
 *
 *	Transfer modes:
 *	%MGACODEC_MODE_I33
 *
 *	%MGACODEC_MODE_VMI_A
 *
 *	%MGACODEC_MODE_VMI_B
 **/
void mgacodec_init(struct mga_codec_s *codec, int mode, int reg, int xor_regs )
{
	codec->mode = mode;
	codec->dstreg = reg;
	codec->xor_regs = xor_regs;

	mgacodec_reset( codec );
}


void
mgacodec_callback(struct mga_codec_s *codec, void (*callback)(struct mga_dev_s *mga, int state))
{
	codec->callback = callback;
}


static void
check_codec_clk(struct mga_codec_s *codec)
{
	if (codec->mga->mga_type == MGA_G400) {

		u32 option2 = 0, old2 = 0;
		if ( pci_read_config_dword( codec->mga->dev, 0x50, &option2 ) != PCIBIOS_SUCCESSFUL ) {
			printk( "mgacodec: error setting clk\n" );
			return;
		}
		dprintk( "CS 50 0x%08x\n", option2 );
		
		old2 = option2;

		/* set codec clk to something sensible; PCI at the moment */
		option2 &= ~0x0F000000;
		/* paranoid: disable writing to seeprom (it's partially
			multiplexed on the codec interface) */
		option2 &= ~0x00000100;
		/* pci=0, syspll=1, mclk=2, agpdll=3 */
		/* dividers: (1<<26) == /2   (1<<27) == /1.5 */
		option2 |= (0 << 24);
		if ( option2 != old2 ) {
			if ( pci_write_config_dword( codec->mga->dev, 0x50, option2 ) != PCIBIOS_SUCCESSFUL ) {
				printk( "mgacodec: error writing CS x50\n" );
			}
			option2 = 0;
			if ( pci_read_config_dword( codec->mga->dev, 0x50, &option2 ) != PCIBIOS_SUCCESSFUL ) {
				printk( "mgacodec: error reading CS x50\n" );
			}
			dprintk( "CS 50 new 0x%08x\n", option2 );
		}
	}
}

/**
 *	mgacodec_reset -	reset the Codec Interface
 *	@codec:	codec instance
 *	
 *	This function resets the Codec Interface to the IDLE state,
 *	disabling transfers.
 *
 *	The order for the 3 CODECCTL bytes should probably be (need to
 *	verify):
 *
 *	1. clear codec general control
 *
 *	2. set recovery width and codecfifoaddr (only at (re)start
 *	transfer?)
 *
 *	3. set codec general control
 *
 *	4. set codecaddr
 *
 *	-inkling
 **/
void mgacodec_reset(struct mga_codec_s *codec)
{
	u8	dstreg;
	int	i;
	unsigned long flags;

	mgacore_printk("mgacodec: starting reset\n");

	spin_lock_irqsave( &codec->state_lock, flags );
	
	if (codec->ctl_shadow & codectransen) {
		dprintk( "mgacodec: stalling codec...\n" );
		mgacodec_and_or_codecctl(codec, ~codectransen, 0 );
		
		for (i=160; i; i--) {
			udelay(1);
			if ((mga_read(codec->mga, VSTATUS) & codec_stalled)) {
				dprintk( "mgacodec: stalled, left=%i\n", i );
				break;
			}
		}
		if (i==0) {
			dprintk( "mgacodec: not stalled in reset\n" );
		}
	}
	codec->transfer = 0;	/* transfer off */
	codec->stalled = 0;	/* stalled off */

	spin_unlock_irqrestore( &codec->state_lock, flags );

		
	dstreg = codec->dstreg ^ codec->xor_regs;

	codec->cmdindex = 0;
	codec->dataindex = 0;
	codec->readindex = 0;
	codec->last_readindex = 0;


	codec->streamindex = 0; /* move pointer back to start */
	
	codec->rwcfa_shadow = ((codec->rwidth & 3)<<4) | (dstreg & 15);

	codec->ctl_shadow = 0;	/* clears on reset */
	codec->misc_shadow = 0;	/* does reset also clear this one? */
	
	check_codec_clk( codec );

	MGA_REGLOCK(codec->mga);

	if (codec->mga->mga_type == MGA_G400) {
	
		mgax_writeb(codec->mga, CODECCTL, 0);
	
		/* I repeat it twice per p. 4-110 of the G400 HB */
		/* could find no corresponding reference in G200 spec
		but it shouldn't hurt anything on the G200, so wtf.
		maybe it's some kind of byte/word transfer issue on AGP. who knows?
		-inkling */
		mgacodec_and_or_codecctl(codec, 0, codecen | codec->mode);
		mgacodec_and_or_codecctl(codec, 0, codecen | codec->mode);
		
		mgax_writeb(codec->mga, CODECCTL+1, codec->rwcfa_shadow);
		
		mga_write(codec->mga, CODECADDR,
			(codec->addr & ~0xFE0003FF ) |
			(CODEC_BUFFER_SIZE == 0x40000 ? 1 : 0) );
	}
	else {
		/* 1. disable codec.  see G200 spec, p 4-89, section 4.9.5 */
		mgax_writeb(codec->mga, CODECCTL, 0);

		/* see G200 spec, p 4-89ff
		2. set codec recovery width BEFORE setting codecen
		and set codecfifoaddr (zoran bitstream register)
		-inkling */
		mgax_writeb(codec->mga, CODECCTL+1, codec->rwcfa_shadow);

		/* set codecen to enable the codec */
		mgacodec_and_or_codecctl(codec, 0, codecen | codec->mode);
		
		/* despite what doc says, this really is needed --lvs */
		mgax_writeb(codec->mga, CODECCTL+1, codec->rwcfa_shadow);

		/* see G200 spec p 3-68: bit0=1=256k buffer, bit0=0=128k buffer */
		mga_write(codec->mga, CODECADDR,
			(codec->addr & ~0xFF0003FF ) |
			(CODEC_BUFFER_SIZE == 0x40000 ? 1 : 0) );
	}
	MGA_REGUNLOCK(codec->mga);

	/* a disable/enable of the codec causes VIEN to loose bits (0x02 and maybe others,
	   0x01 stays alive). re-enable all registered ints:
	   --lvs
	*/
	mga_check_irqs(codec->mga);
	
	mgacore_printk("mgacodec: reset codec done\n");
}


/*
 * calculates how many bytes are left to write new data in the
 * circular buffer
 *
 */ 
u32 mgacodec_buffer_free(struct mga_codec_s *codec)
{
	int free; /* int32 math ok, it's not greather than +/- 2G */

	free = mgacodec_hardptr(codec) - codec->streamindex;
	if (free <= 0) 
		free += CODEC_BUFFER_SIZE;
	return free;
}

/* stops the codec interface automatic bitstream transfer
   this stalls any transfer in progress, disables the buffer level IRQ
   and resets the codec. it also clears transfer and stalled flags.
   this is needed to reset the pointers so you don't get garbage playing
   two different mpegs in a row without close/reopen mgadvd device.
   -inkling */
int mgacodec_stop_transfer(struct mga_codec_s *codec)
{
	dprintk("mgacodec: stopping codec transfer\n");

	mgacodec_stall_transfer(codec);
	if (codec->callback) 
		(codec->callback)(codec->mga, MGACODEC_STOP);

	mgacodec_reset(codec);

	dprintk("mgacodec: %s transfer stopped\n",
		  (codec->transfer==CODEC_SEND_TRANSFER)
			? "SEND":"RECEIVE");
	return 1;
}


/* program the codec interface to raise an interrupt once a certain 
   byte offset is reached by the bitstream transfer engine
*/
void mgacodec_set_transfer_limit(struct mga_codec_s *codec, u32 hostptr)
{
	/* CODECHOSTPTR must be on an 8 _dword_ boundary (p. 4-248 G100)
	   it is the offset _inside_ the buffer of a _dword_
	*/
	hostptr >>= 2;
	hostptr &= ~0xFFFF0007; /* top 16 bits must be zero G200 spec p 3-72 */

	mga_write(codec->mga, CODECHOSTPTR, hostptr);
}


u32 mgacodec_hardptr(struct mga_codec_s *codec)
{
	return (mga_read(codec->mga, CODECHARDPTR) & 0xffff) << 2;
}


u32 mgacodec_lcode(struct mga_codec_s *codec)
{
	return (mga_read(codec->mga, CODECLCODE) & 0xffff) << 2;
}


u32 mgacodec_bitstream(struct mga_codec_s *codec)
{
	return codec->streamindex;
}

/* re/set codecctl bits in copy and use copy to set codecctl */
void mgacodec_and_or_codecctl(struct mga_codec_s *codec, u8 and, u8 or)
{
	unsigned long flags;

	spin_lock_irqsave( &codec->lock, flags );

	codec->ctl_shadow &= and;
	codec->ctl_shadow |= or;

	/* write byte NOT word because according to the G200 spec, p 4-89,
	   writing to codec recovery/codecfifoaddr byte can cause corruption
	   during data transfer. this agress with the working april code.
	   see mgacodec_reset for setting the recovery width at CODECCTL+1
	   -inkling */

	mgax_writeb(codec->mga, CODECCTL, codec->ctl_shadow);

	codec->ctl_shadow &= ~cmdexectrig;
	
	spin_unlock_irqrestore( &codec->lock, flags );
}



/**
 *	mgacodec_and_or_miscctl -	twiddle MISCCTL
 *	@codec:	codec instance
 *	@and:	AND operand
 *	@or:	OR operand
 *
 * 	This function twiddles the MISCCTL part of the codec interface
 *	to reset the actual CODEC chip connected to the MISC lines of
 *	the interface.
 *
 * 	QUESTION: should this be called AFTER recovery width setting?
 *	G200 docs say to set recovery width BEFORE enabling codec.
 *	this code enables the codec.
 *	-inkling
 **/
void mgacodec_and_or_miscctl(struct mga_codec_s *codec, u8 and, u8 or)
{
	unsigned long flags;

	spin_lock_irqsave( &codec->lock, flags );

	/* see G200 spec p 4-90ff for compression information
	   see G200 spec p 4-93ff for decompression information
	   we must be in IDLE (or COMP/DECOMP) for this to work

	   decompression:
	   p. 4-92 states that MISC<2> is connected to EOI (?) when the
	   CODEC Interface is enabled.
	   So it seems that we can bang on these 2 bits too...
	*/

	/* we MUST either verify we are in IDLE (or COMP/DECOMP), or force
	   the right state before doing this. see G200 spec p 4-89ff
	   we should send a 0 to the low byte of codecctl (to stop the codec;
	   optional but GOOD IDEA) then send a 1 to the low byte of codecctl
	   (to enable the codec) and then send the misctl bits to the high
	   byte of codecctl. the procedure is detailed in G200 spec p 4-94ff
	   -inkling */

#ifdef	NEVER
	/* this follows the G200 spec p 4-89ff
	   disable (spec says optional, but should force IDLE state)
	   then enable the codec, then program the MISCCTL byte.
	   also, update ctl shadow with new value since we changed CODECCTL+0
	   and update the misc shadow too.
	   -inkling */
	mgax_writeb(codec->mga, CODECCTL, 0);
	mgax_writeb(codec->mga, CODECCTL, 1);
	codec->ctl_shadow = 1;
#endif
	codec->misc_shadow &= and;
	codec->misc_shadow |= or;
	mgax_writeb(codec->mga, CODECCTL+3, codec->misc_shadow);
	
	spin_unlock_irqrestore( &codec->lock, flags );
}


u8 mgacodec_get_miscctl(struct mga_codec_s *codec)
{
	return codec->misc_shadow;
}

/**
 *	mgacodec_set_misc2 -	program the MISC<2> line
 *	@codec:	codec instance
 *	@param:	0 to disable the output driver, 1 to enable the output driver
 *
 * 	Set the mode of the MISC<2> line (input or output).
 *
 *	See:	p. 3-325 (361) of the G200 spec
 **/
int mgacodec_set_misc2(struct mga_codec_s *codec, u8 param)
{
	u8 b;

	MGA_REGLOCK(codec->mga);
	/* codec should be/is enabled when calling us
	mgacodec_and_or_codecctl(codec, 0xFF, codecen | codec->mode);
	mgacodec_and_or_miscctl(codec, ~param, 0);
	mgacodec_and_or_codecctl(codec, ~codecen, 0);
	*/
	mga_indirect_and_or(codec->mga, XGENIOCTRL, ~0x40, 0x00);
	b = mga_indirect_readb(codec->mga, XGENIOCTRL);
	MGA_REGUNLOCK(codec->mga);
	
	if (b & 0x40) {
		printk("mgacodec: error setting misc2 line\n");
		return 1;
	}
	return 0;
}

/**
 *	mgacodec_set_ddc2 -	program the DDC<2> line
 *	@codec:	codec instance
 *	@param:	0 to disable the output driver, 1 to enable the output driver
 *
 * 	Set the mode of the DDC<2> line (input or output).
 *
 * 	Rationale: p. 5-69 (389) and p. 5-92 (392) of the G_1_00 spec
 * 	The DDC<2> line is connected to the HRDY line for a DVD addon board 
 **/
int mgacodec_set_ddc2(struct mga_codec_s *codec, u8 param)
{
	u8 b;

	MGA_REGLOCK(codec->mga);
	/* codec should be/is enabled when calling us
	mgacodec_and_or_codecctl(codec, 0xFF, codecen | codec->mode);
	mgacodec_and_or_miscctl(codec, ~param, 0);
	mgacodec_and_or_codecctl(codec, ~codecen, 0);
	*/
	mga_indirect_and_or(codec->mga, XGENIOCTRL, ~0x04, 0x00);
	b = mga_indirect_readb(codec->mga, XGENIOCTRL);
	MGA_REGUNLOCK(codec->mga);

	if (b & 0x04) {
		mgacore_printk("mgacodec: error setting ddc2 line\n");
		return 1;
	} 
	return 0;
}


void mgacodec_begin_commands(struct mga_codec_s *codec)
{
	spin_lock( &codec->cmd_lock );
}


void mgacodec_end_commands(struct mga_codec_s *codec)
{
	spin_unlock( &codec->cmd_lock );
}



/**
 * 	mgacodec_push_readb -	add a read instruction (8 bits)
 *	@codec:	codec instance
 *	@reg:	register to read
 *
 *	Add a read instruction to the command buffer.
 *	To actually read a register, you must call
 *	mgacodec_exec_commands() and mgacodec_get_word() after this.
 **/
void mgacodec_push_readb(struct mga_codec_s *codec, u8 reg)
{
	reg ^= codec->xor_regs;

	codec->last_cmdindex = codec->cmdindex;

	codec->cmdbuffer[codec->cmdindex++] = (1<<15) | (reg<<10) | 1;
	codec->cmdbuffer[codec->cmdindex++] = 0;
	codec->readindex += 4;
}

/**
 * 	mgacodec_push_readw -	add a read instruction (16 bits)
 *	@codec:	codec instance
 *	@reg:	register to read
 *
 *	Add a read instruction to the command buffer.
 *	To actually read a register, you must call
 *	mgacodec_exec_commands() and mgacodec_get_word() after this.
 **/
void mgacodec_push_readw(struct mga_codec_s *codec, u8 reg)
{
	reg ^= codec->xor_regs;
	
	codec->last_cmdindex = codec->cmdindex;

	codec->cmdbuffer[codec->cmdindex++] = (1<<15) | (reg<<10) | 2;
	codec->cmdbuffer[codec->cmdindex++] = 0;
	codec->readindex += 4;
}

/**
 * 	mgacodec_push_writeb -	add a write instruction (8 bits)
 *	@codec:	codec instance
 *	@reg:	register to write
 *	@data:	byte to write
 *
 *	Add a write instruction to the command buffer.
 *	To actually write the data, you must call
 *	mgacodec_exec_commands() after this.
 *
 *	See also:	mgacodec_push_readb(),
 *	mgacodec_push_writeb(), mgacodec_exec_commands()
 **/
void mgacodec_push_writeb(struct mga_codec_s *codec, u8 reg, u8 data)
{
	reg ^= codec->xor_regs;
	
	codec->last_cmdindex = codec->cmdindex;

	codec->cmdbuffer[codec->cmdindex++] = 1;
	codec->cmdbuffer[codec->cmdindex++] = 0;

	codec->cmdbuffer[codec->cmdindex++] = (reg << 8) | data;
	codec->cmdbuffer[codec->cmdindex++] = 0;
}

/**
 * 	mgacodec_push_writew -	add a write instruction (16 bits)
 *	@codec:	codec instance
 *	@reg:	register to write
 *	@data:	word to write
 *
 *	Add a write instruction to the command buffer.
 *	To actually write the data, you must call
 *	mgacodec_exec_commands() after this.
 *
 *	See also:	mgacodec_push_readw(),
 *	mgacodec_push_writew(), mgacodec_exec_commands()
 *
 *	Notes: this function is tuned to transfer words to a
 *	big-endian chip (like the ZR367x0 chipset).
 *
 *	Fixme: Only i386 has been tested so far.
 **/
void mgacodec_push_writew(struct mga_codec_s *codec, u8 reg, u16 data)
{
	reg ^= codec->xor_regs;

	codec->last_cmdindex = codec->cmdindex;

	codec->cmdbuffer[codec->cmdindex++] = 2;
	codec->cmdbuffer[codec->cmdindex++] = 0;

	codec->cmdbuffer[codec->cmdindex++] = (CODEC_WRITE(reg, (data & 0x00FF))<<16)
		| CODEC_WRITE(reg, ((data>>8) & 0xFF));
	codec->cmdbuffer[codec->cmdindex++] = 0;
}

/**
 * 	mgacodec_push_pause -	add a pause in the command stream
 *	@codec:	codec instance
 *
 *	This function sets the pause bit in the last 'pushed
 * 	command stream, as described in the Matrox spec. Once the
 *	Codec Interface see the 'pause' bit it will automtically
 *	enable data transfer (read or write).
 *
 *	See also:	mgacodec_push_readb/w(),
 *	mgacodec_push_writeb/w(), and mgacodec_exec_commands()
 **/
void mgacodec_push_pause(struct mga_codec_s *codec)
{
	codec->cmdbuffer[codec->last_cmdindex] |= (1<<8);
}

/**
 *	mgacodec_get_byte -	get the result of the last commands (8 bits)
 *	@codec:	codec instance
 *
 *	This function reads one byte in the data buffer returned by
 *	previous read instructions. Successive calls dequeue the bytes
 *	received in the data buffer.
 *
 *	Return value:	a byte
 *
 *	See also:	mgacodec_push_readb(),
 *	mgacodec_push_writeb(), mgacodec_exec_commands()
 **/
u8 mgacodec_get_byte(struct mga_codec_s *codec)
{
	u8 data;

	if ( codec->dataindex+4 > codec->last_readindex ) {
		dprintk( "get_byte: no data (%i %i)\n", codec->last_readindex, codec->dataindex );
	}
	data = codec->databuffer[codec->dataindex] & 0x00FF;
	codec->dataindex += 4;

	return data;
}

/**
 *	mgacodec_get_word -	get the result of the last commands (16 bits)
 *	@codec:	codec instance
 *
 *	This function reads one word in the data buffer returned by
 *	previous read instructions. Successive calls dequeue the word
 *	received in the data buffer.
 *
 *	Return value:	a word
 *
 *	See also:	mgacodec_push_readw(),
 *	mgacodec_push_writew(), mgacodec_exec_commands()
 *
 *	Notes: this function is tuned to transfer words to a
 *	big-endian chip (like the ZR367x0 chipset).
 *
 *	Fixme: Only i386 has been tested so far.
 **/
u16 mgacodec_get_word(struct mga_codec_s *codec)
{
	u16 data;

	if ( codec->dataindex+4 > codec->last_readindex ) {
		dprintk( "get_word: no data (%i %i)\n", codec->last_readindex, codec->dataindex );
	}
	data = __swab16(codec->databuffer[codec->dataindex]);
	codec->dataindex += 4;

	return data;
}


static void mgacodec_cmdcmpl(struct mga_dev_s *mga, struct pt_regs * regs)
{
	mga->codec->cmdq_hascmpl = 1;
	wake_up_interruptible(&mga->codec->cmdq);
}

/**
 *	mgacodec_exec_commands - 	executes queued commands
 *	@codec:	codec instance
 *
 * 	This function executes the commands queued in the command
 *	buffer by previous mgacodec_push_read() (or _write()) calls.
 *
 *	It takes care of suspending any bitstream transfer that may
 *	take place at the time it is called, without loosing datas (it
 *	first asks the codec interface to stall transfer and waits for
 *	it to complete). This function also automatically restarts an
 *	interrupted transfer after commands have been executed.
 *
 */
int mgacodec_exec_commands(struct mga_codec_s *codec, u8 and, u8 or)
{
	size_t cmdbytes;
	size_t readbytes;

	mgacodec_stall_transfer(codec);
	
	cmdbytes = codec->cmdindex * sizeof(codec->cmdbuffer[0]);
	if ( cmdbytes > CODEC_CMDAREA_SIZE ) {
		codec->dataindex = 0;
		codec->cmdindex = 0;
		printk( KERN_ERR "codec: exec_commands cmd overflow\n" );
		return 1;
	}
	
	readbytes = codec->readindex*sizeof(codec->databuffer[0]);
	if ( readbytes > CODEC_ARGAREA_SIZE ) {
		codec->dataindex = 0;
		codec->cmdindex = 0;
		printk( KERN_ERR "codec: exec_commands read overflow\n" );
		return 1;
	}

	/* activate the stop bit */
	codec->cmdbuffer[codec->last_cmdindex] |= (1<<9);

	/* copy command buffer into video memory */
	memcpy_toio((u8*)codec->mga->fb + codec->addr + CODEC_BUFFER_SIZE,
			codec->cmdbuffer, cmdbytes);

	/* invalidate data buffer for return values */
	memset_io((u8*)codec->mga->fb + codec->addr + CODEC_BUFFER_SIZE + CODEC_CMDAREA_SIZE, 0xDD,
		readbytes );

	mga_flush( codec->mga );

	codec->cmdq_hascmpl = 0;

	/* exec cmds loaded in cmdbuffer */
	mgacodec_and_or_codecctl(codec, and, or | cmdexectrig );


	codec->dataindex = 0;
	codec->cmdindex = 0;
	codec->last_cmdindex = 0;
	codec->last_readindex = codec->readindex;
	codec->readindex = 0;

	if (signal_pending(current)) {
		dprintk( "cmdexec while signal pending\n" );
		udelay( 25 );
	} else {
		/* sleep waiting for command complete with timeout */
		interruptible_sleep_on_timeout(&codec->cmdq, HZ/100 + 1);
		if (signal_pending(current)) {
			printk("mgacodec: command exec interrupted hascmpl %d\n", 
				codec->cmdq_hascmpl);
			/* keep running */
			udelay( 25 );
		}
	}

	if (!codec->cmdq_hascmpl) {
		printk( MGA_ERR "mgacodec: command exec wakeup without hascmpl\n");
	}
	
	if ( readbytes ) {
		mga_flush( codec->mga );
		memcpy_fromio(codec->databuffer, (u8*)codec->mga->fb + codec->addr + 
		 	CODEC_BUFFER_SIZE + CODEC_CMDAREA_SIZE,
			readbytes );
	}
	
	if (codec->stalled)
		mgacodec_start_transfer( codec, codec->transfer );

	return 0;
}


int mgacodec_get_transfer(struct mga_codec_s *codec)
{
	return codec->transfer;
}


int mgacodec_is_stalled(struct mga_codec_s *codec)
{
	return codec->stalled;
}


/* puts the codec interface in bitstream transfer mode
   sets transfer flag, clears stalled flag, and sets direction
*/
int mgacodec_start_transfer(struct mga_codec_s *codec, int direction)
{
	u8 ctl_mask = 0xff;
	u8 ctl = codecen | codectransen;
	unsigned long flags;

	spin_lock_irqsave( &codec->state_lock, flags );

	if (codec->transfer)
	{
	/* check if we need to restart a transfer that has been stalled
	   because of a codec buffer level IRQ occurring
	   -inkling */
		if (codec->stalled)
		{
			/* keep current ctl (codectransen should be 0) */
			codec->stalled = 0;
			
			dprintk2("mgacodec: %s transfer re-starting\n",
			  (codec->transfer==CODEC_SEND_TRANSFER)
				? "SEND":"RECEIVE");

			mgacodec_and_or_codecctl(codec, ctl_mask, ctl);

		} else {
			dprintk("mgacodec: %s transfer already started!\n",
			  (codec->transfer==CODEC_SEND_TRANSFER)
				? "SEND":"RECEIVE");
		}
	} else {	
	/* no transfer has started. start it now.
	   set the transfer direction and reset stalled flag.
	   codec buffer level IRQ will set the stalled flag later.
	   -inkling */

		/* set which direction we're going but i don't see
		   where there is a direction setting in the G200 spec
		   so direction is apparently useless to start/stall/stop
		   -inkling */
		codec->transfer = direction;
		codec->stalled = 0;
	
		if (codec->transfer == CODEC_RECV_TRANSFER)
			ctl |= codecdatain;

		dprintk("mgacodec: %s transfer starting\n",
		  (codec->transfer==CODEC_SEND_TRANSFER)
			? "SEND":"RECEIVE");

		/*
		 * Need for zr36060 to step start line before
		 * Codec transfer is enabled.
		 */
		if (codec->callback)
			(codec->callback)(codec->mga, MGACODEC_START);

		mgacodec_and_or_codecctl(codec, ctl_mask, ctl);
	}

	spin_unlock_irqrestore( &codec->state_lock, flags );

	return 0;
}


#define SENDSTREAM_HEAD 4096
/**
 *	mgacodec_send_bitstream -	initiates bitstream transfer
 *	@codec:		codec instance
 *	@bitstream:	block to send
 *	@size:		size of the block
 *
 *	This function initiates bitstream transfer to the decoder
 *	chip. It writes the bitstream to the Codec Interface buffer
 *	(using the interrupt handler). More specifically, what we do
 *	here is :
 *
 *	    - check if there is room left to write or _wait_ until
 *	      there is;
 *
 *	    - write a block of data;
 *
 *	    - tell the interrupt handler to stall the transfer if the
 *	      codec interface runs past the last written block.
 */
int mgacodec_send_bitstream(struct mga_codec_s *codec, const u8 *bitstream, unsigned int size)
{
	u32 buf_free;
	u32 buf_limit;
	u32 headsize;
	u32 filled;
	unsigned long flags;

	headsize = size + SENDSTREAM_HEAD;/* headroom */



	/*
	  Codec Bitstream Buffer
	  [..........................................................]
	            ^                ^         ^
		    hardptr          hostptr   streamindex

		  o  hardptr is driven by the hardware: it's a pointer
		     to the qword currently sent to the decoder
		  o  hostptr is handled by our driver: it triggers an
		     IRQ to stall the codec when hardptr reaches it
		  o  streamindex is handled by our driver: it marks
		     the point in the codec buffer where we're
		     currently writing
	*/

	/* if the transfer has not yet started, prefill the buffer
	    until it is nearly full before we start the transfer
	    -inkling */
	buf_free = mgacodec_buffer_free(codec);

	if ((codec->transfer == CODEC_NO_TRANSFER) || codec->stalled) {
	
		if (buf_free > CODEC_BUFFER_SIZE/ 2) {
			if ( buf_free <= size ) {
				printk( KERN_ERR "mgacodec: writing too much at once\n" );
			}
			mgacodec_set_transfer_limit(codec, codec->streamindex);
			mgacodec_copy_bitstream(codec, bitstream, size);
			/* buffer is not full enough yet, keep filling */
			return 0;
		} else {
			printk("mgacodec: start_transfer; limit at %d\n", codec->streamindex);
			mgacodec_set_transfer_limit(codec, codec->streamindex);
			mgacodec_start_transfer(codec, CODEC_SEND_TRANSFER);
		}
	}

	/* the transfer has started, check if we can add more data.
	   if we can't add more data, we will have to wait for the
	   codec buffer level IRQ to indicate that some space is free.
	   -inkling */

	/* set busy flag so we don't stall while waiting for room in buffer */
	codec->send_busy = 1;

	/* using buflvl irq for two purposes:
	     -stall on underrun
	     -wake-up when space available
	*/
	/* set point of irq at position where enough room available */
	buf_limit = (codec->streamindex + headsize + 2*8*4) & (CODEC_BUFFER_SIZE-1);
	mgacodec_set_transfer_limit(codec, buf_limit);

	if ( buf_free <= headsize ) {
	/* we have to wait for some space to be freed up by waiting
	   until the codec reaches buffer level IRQ limit and stalls.
	   set the interrupt to happen, then wait for it.
	   -inkling */

	   	int rv;

		rv = wait_event_interruptible(codec->outq, (buf_free=mgacodec_buffer_free(codec))>headsize );
		/* abort if we received any signal during this wait */
		if (rv) {
			mgacore_printk("mgacodec: wait interrupted by signal\n");
			codec->send_busy = 0;
			return rv;
		}
	}
        
	/* write the data to the buffer and adjust the buffer level IRQ
	   to stall the codec once it is reached. this keeps any stalls
	   in the feed stream from corrupting the playback.

	   this provides the very much desired benefit of allowing the
	   codec to gracefully handle hardware induced stalls in the feed
	   stream (dvd sector read pause at layer change or task switching.)

	   -inkling */

	mgacodec_copy_bitstream(codec, bitstream, size);

	/* ATOMIC start; needed for calc of good buf_limit */
	local_irq_save(flags); // Ughh..  Needs fixing.
	
	filled = CODEC_BUFFER_SIZE - mgacodec_buffer_free(codec);
	if ( filled >= 32768 )
		buf_limit = (codec->streamindex - (32768-4096)) & (CODEC_BUFFER_SIZE-1);
	else
		buf_limit = (codec->streamindex - (filled/2)) & (CODEC_BUFFER_SIZE-1);
	mgacodec_set_transfer_limit(codec, buf_limit);
	
	codec->send_busy = 0;
	
	local_irq_restore(flags);  /* ATOMIC end */

	return 0;	
}


/* non-blocking irq safe version */
int mgacodec_send_bitstream_irq(struct mga_codec_s *codec, const u8 *bitstream, unsigned int size)
{
	u32 buf_free;

	buf_free = mgacodec_buffer_free(codec);

	if ( buf_free < size ) {
		printk("mgacodec: send overrun free %x < size %x\n", buf_free, size );
		return 1;
	}

	mgacodec_copy_bitstream(codec, bitstream, size);
	mgacodec_set_transfer_limit(codec, codec->streamindex);

	if (codec->transfer == CODEC_NO_TRANSFER) {
		if (buf_free < CODEC_BUFFER_SIZE/ 2) {
			/* buffer is full enough */
			mgacodec_start_transfer(codec, CODEC_SEND_TRANSFER);
		}
	}
	return 0;
}


int mgacodec_recv_bitstream(struct mga_codec_s *codec, u8 *addr, int start, int end, int limit)
{
	int	transfer;
	int	chunk;

	transfer = 0;

	mga_flush( codec->mga );

	/*
	 * Check to see if transfer past end of the codec ring. If
	 * so transfer from start to end of ring and adjust 
	 * vars for data removed.
	 */
	if (start > end) {
		chunk = CODEC_BUFFER_SIZE - start;
		if (chunk > limit)
			chunk = limit;
		memcpy_fromio(addr, (u8*)codec->mga->fb + codec->addr + 
			start, chunk);
		limit -= chunk;
		transfer += chunk;
		if ( limit == 0 )
			return transfer;
		addr += chunk;
		start = 0;
	}

	chunk = end - start;
	if (chunk > limit)
		chunk = limit;

	memcpy_fromio(addr, (u8*)codec->mga->fb + codec->addr + 
		start, chunk);
	transfer += chunk;

	return transfer;
}

/**
 *	mgacodec_buflevel_irq -	handle BLVL interrupts
 *	@mga:	mga instance
 *	@regs:	general registers passed to an interrupt handler
 *
 *	This routine handles BLVL interrupts. It is called each time
 *	the Codec Interface transfer engine has reached a programmed
 *	limit. This routine just stalls the bitstream transfer to
 *	avoid data loss and bitstream corruption and wakes up queue
 *	tasks. For example, if the Codec Interface works in
 *	SEND_TRANSFER mode (decompression), then this routine will
 *	signal the 'outq' wait queue.
 *
 *	See also:	mga_enable_irq() to install the routine
 *
 *	Note:	this routine is installed by default when the
 *	mgacodec_send_bitstream() function is called repeatedly. This
 *	will change to accommodate the recv_bitstream variant.
 */
void mgacodec_buflevel_irq(struct mga_dev_s *mga, struct pt_regs *regs)
{
	struct mga_codec_s *codec = mga->codec;

	if (!codec->send_busy) {
		dprintk("codec_blvl callback %p\n", codec->callback);

		if (codec->callback)
			(codec->callback)(codec->mga, MGACODEC_STOP);
	}

	if (codec->transfer == CODEC_SEND_TRANSFER)
		wake_up_interruptible(&codec->outq);
	if (codec->transfer == CODEC_RECV_TRANSFER)
		wake_up_interruptible(&codec->inq);
}

/**
 *	mgacodec_poll -	handle poll syscall
 *	@codec:	codec_instance
 *	@file:	see &file_operations
 *	@wait:	see &file_operations
 *	@read_threshold:	minimum number of bytes to consider
 *				that read will succeed
 *	@write_threshold:	minimum number of bytes to consider
 *				that write will succeed
 *
 *	mgacodec_poll() is a helper function for driver implementing
 *	the 'poll' file operation. The threshold parameters help
 *	ensure that a subsequent 'write' system call won't block
 *	because there isn't enough space left.
 *
 **/
unsigned int mgacodec_poll(struct mga_codec_s *codec, struct file *file, poll_table *wait,
			   u32 read_threshold, u32 write_threshold)
{
	unsigned int mask;

	poll_wait(file, &codec->inq, wait);
	poll_wait(file, &codec->outq, wait);
	mask = 0;
	if (mgacodec_buffer_free(codec) > (write_threshold + SENDSTREAM_HEAD))
                mask |= POLLOUT | POLLWRNORM;

	/* FIXME: add more stuff here when compressing video */

	return mask;
}


/* set the vsyncq active (wakes up vsync_sleep thread if there is one) */
static void
mgacodec_vsync_irq(struct mga_dev_s *mga, struct pt_regs * regs)
{
	mga->codec->vsyncq_hasirq = 1;
	wake_up_interruptible(&mga->codec->vsyncq);
}

/* sleeps until vinvsync flag, then returns 0 if not signal kill
 *
 * have to call this repeatedly from i34_poll_status thread but it
 * does go to sleep until the next vinvsync, instead of eating cpu.
 * when it wakes up, i34(i33 too?) can check the status (hopefully) -ink
 */
int
mgacodec_vsync_sleep(struct mga_codec_s *codec)
{
	do
	{
	/* clear flag and check for signal to break us out */
		codec->vsyncq_hasirq = 0; 
		if (signal_pending(current)) {
			dprintk( "vsync poll while signal pending\n" );
			udelay( 25 );
			break;
		} else {
		/* sleep waiting for vinvsync with timeout of 2 seconds */
			interruptible_sleep_on_timeout(&codec->vsyncq, HZ + 1);
			if (signal_pending(current)) {
				printk("mgacodec: vsync sleep got hasirq %d\n",
					codec->vsyncq_hasirq); 
			/* keep running */
			udelay( 25 );
			}
		}
		if (!codec->vsyncq_hasirq) { 
			printk( MGA_ERR "mgacodec: vsync_sleep wakeup w/o hasirq\n");
		} 

        /* return to calling thread if vsync has woken us up */
	if (codec->vsyncq_hasirq) return 0;

	/* otherwise go back to sleep */
    } while (1);

/* signal killed me */
return 1;
}
